"""
SiliconLife Eyeflow
Class for access files synchronized with azure cloud storage

Author: Gabriel Melo
"""

# pip3 install azure-storage-blob
from azure.storage.blob import BlobServiceClient
from azure.core.exceptions import ResourceNotFoundError, HttpResponseError
#----------------------------------------------------------------------------------------------------------------------------------

class Connector():
    """
    Azure blobs access
    """

    def __init__(self, **kwargs):
        self.account_url = kwargs["account_url"]
        self.account_key = kwargs["account_key"]
        self.client = BlobServiceClient(
            account_url=self.account_url,
            credential=self.account_key
            )


    def _get_container_client(self, folder):
        return self.client.get_container_client(container=folder)


    def _get_blob_client(self, folder, filename):
        return self.client.get_blob_client(container=folder, blob=filename)


    def list_files(self, folder, resource_id=None, prefix='', sufix='', full_path=False):
        container_client = self._get_container_client(folder)
        file_list = []
        for blob in container_client.list_blobs(name_starts_with=resource_id if resource_id is not None else ''):
            if full_path:
                blob_name = blob['name']
            else:
                blob_name = blob['name'].split('/').pop()

            if blob_name.startswith(prefix) and blob_name.endswith(sufix):
                file_list.append(blob_name)

        return file_list


    def list_files_info(self, folder, resource_id=None, prefix='', sufix='', full_path=False):
        """
        Returns a list with info of all files from folder
        [{"filename": string,
          "creation_date": timestamp,
          "modified_date": timestamp,
          "file_size": int}
        ]
        """
        container_client = self._get_container_client(folder)
        file_list = []
        for blob in container_client.list_blobs(name_starts_with=resource_id + '/' if resource_id is not None else ''):
            if full_path:
                blob_name = blob['name']
            elif resource_id is not None:
                blob_name = blob['name'].replace(resource_id + '/', '', 1)
            else:
                blob_name = blob['name']

            if blob_name.startswith(prefix) and blob_name.endswith(sufix):
                file_info = {"filename": blob_name,
                             "creation_date": blob["creation_time"],
                             "modified_date": blob["last_modified"],
                             "file_size": blob["size"]
                            }

                file_list.append(file_info)

        return file_list


    def is_file(self, folder, filename, resource_id=None):
        """
        Returns true if file exists
        """
        if resource_id is not None:
            blob_name = f"{resource_id}/{filename}"
        else:
            blob_name = filename

        try:
            blob_client = self._get_blob_client(folder, blob_name)
            blob_client.get_blob_properties()
            return True
        except ResourceNotFoundError:
            return False


    def get_file_info(self, folder, filename, resource_id=None):
        """
        Returns file info in a dict
        {"filename": string,
         "creation_date": timestamp,
         "modified_date": timestamp,
         "file_size": int
        }
        """
        if resource_id is not None:
            blob_name = f"{resource_id}/{filename}"
        else:
            blob_name = filename

        blob_client = self._get_blob_client(folder, blob_name)
        blob = blob_client.get_blob_properties()
        blob_name = blob['name'].split('/').pop()

        blob_info = {"filename": blob_name,
                     "creation_date": blob["creation_time"],
                     "modified_date": blob["last_modified"],
                     "file_size": blob["size"]
                    }

        return blob_info


    def download_file(self, folder, filename, resource_id=None):
        if resource_id is not None:
            blob_name = f"{resource_id}/{filename}"
        else:
            blob_name = filename

        blob_client = self._get_blob_client(folder, blob_name)
        try:
            download_stream = blob_client.download_blob()
        except Exception as err:
            raise Exception(f'Error downloading {filename}: {err}')
        else:
            data = download_stream.readall()
            return data


    def upload_file(self, folder, filename, data, resource_id=None):
        if not isinstance(data, bytes):
            raise Exception(f'Data should be byte type, {type(data)} detected.')

        if resource_id is not None:
            blob_name = f"{resource_id}/{filename}"
        else:
            blob_name = filename

        blob_client = self._get_blob_client(folder, blob_name)
        try:
            blob_client.upload_blob(data=data)
            blob_client.get_blob_properties()
        except Exception as err:
            raise Exception(f'Error uploading {filename}: {err}')
        else:
            # print(f'Successfully uploaded {filename}')
            pass


    def delete_file(self, folder, filename, resource_id=None):
        if resource_id is not None:
            blob_name = f"{resource_id}/{filename}"
        else:
            blob_name = filename

        blob_client = self._get_blob_client(folder, blob_name)
        try:
            blob_client.delete_blob()
        except Exception as err:
            raise Exception(f'Error deleting {filename}: {err}')
        else:
            # print(f'Successfully deleted {blob_name}')
            pass
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
WS-Edge client functions

Author: Alex Sobral de Freitas
"""

import os
from subprocess import call
from pathlib import Path
import datetime
import json
import cv2
import requests
import tarfile
import jwt

from eyeflow_sdk.log_obj import CONFIG, log
from eyeflow_sdk.img_utils import resize_image_scale
# ---------------------------------------------------------------------------------------------------------------------------------


def get_list_files_info(folder):
    file_list = []
    for filename in os.listdir(folder):
        if os.path.isfile(os.path.join(folder, filename)):
            file_list.append(filename)
        elif os.path.isdir(os.path.join(folder, filename)):
            subfolder = os.path.join(folder, filename)
            subfolder_files = get_list_files_info(subfolder)
            for filename in subfolder_files:
                filename = os.path.join(os.path.split(subfolder)[1], filename)
                file_list.append(filename)

    return file_list
#----------------------------------------------------------------------------------------------------------------------------------


def download_file(url, local_filename):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()

        if os.path.isfile(local_filename):
            os.remove(local_filename)

        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment 'if' and set chunk_size parameter to None.
                #if chunk:
                f.write(chunk)
# ---------------------------------------------------------------------------------------------------------------------------------


def get_dataset(app_token, dataset_id):
    try:
        log.info(f"Get dataset {dataset_id}")

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        response = requests.get(f"{endpoint}/dataset/{dataset_id}", headers=msg_headers)

        if response.status_code != 200:
            log.error(f"Failing get dataset: {response.json()}")
            return None

        dataset = response.json()
        if dataset["dataset_parms"]:
            return dataset
        else:
            log.warning(f"Failing get dataset: {response.json()}")
            return None

    except requests.ConnectionError as error:
        log.error(f'Failing get dataset_id: {dataset_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing get dataset_id: {dataset_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing get dataset_id: {dataset_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_flow(app_token, flow_id):
    try:
        log.info(f"Get flow {flow_id}")

        if not Path(CONFIG["flow_folder"]).is_dir():
            Path(CONFIG["flow_folder"]).mkdir(parents=True, exist_ok=True)

        local_cache = os.path.join(CONFIG["flow_folder"], flow_id + '.json')

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        response = requests.get(f"{endpoint}/flow/{flow_id}", headers=msg_headers)

        if response.status_code != 200:
            log.error(f"Failing get flow from edge: {response.json()}")

            if os.path.isfile(local_cache):
                with open(local_cache) as fp:
                    flow = json.load(fp)

                return flow

            return None

        flow = response.json()["flow"]
        if "_id" in flow:

            if os.path.isfile(local_cache):
                os.remove(local_cache)

            with open(local_cache, 'w') as fp:
                json.dump(flow, fp, default=str)

            return flow
        else:
            log.warning(f"Failing get flow: {response.json()}")
            return None

    except requests.ConnectionError as error:
        log.error(f'Failing get flow_id: {flow_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing get flow_id: {flow_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing get flow_id: {flow_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_model(app_token, dataset_id, model_folder):
    local_doc = None
    try:
        log.info(f"Get model {dataset_id}")

        folder_path = Path(model_folder + '/' + dataset_id)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        local_cache = os.path.join(model_folder, dataset_id + '.json')
        if os.path.isfile(local_cache):
            with open(local_cache) as fp:
                local_doc = json.load(fp)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        url = f"{endpoint}/published-model/{dataset_id}/"
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        payload = {"download_url": False}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model: {response.json()}")
            return None

        model_doc = response.json()
        if local_doc and model_doc["date"] == local_doc["date"]:
            return local_doc

        payload = {"download_url": True}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model: {response.json()}")
            return None

        model_doc = response.json()

        dest_filename = os.path.join(model_folder, dataset_id + ".tar.gz")
        download_file(model_doc["download_url"], dest_filename)

        # expand_file
        call([
            "tar",
            "-xzf", dest_filename,
            "--directory", folder_path
        ])

        os.remove(dest_filename)

        if os.path.isfile(local_cache):
            os.remove(local_cache)

        with open(local_cache, 'w') as fp:
            json.dump(model_doc, fp, default=str)

        return model_doc

    except requests.ConnectionError as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get model dataset_id: {dataset_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get model dataset_id: {dataset_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing get model dataset_id: {dataset_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def upload_model(
    app_token,
    dataset_id,
    model_info,
    model_folder,
    train_id,
    train_info,
    hist_folder
):
    try:
        model_filename = os.path.join(model_folder, dataset_id + ".tar.gz")
        if os.path.isfile(model_filename):
            os.remove(model_filename)

        folder_path = os.path.join(model_folder, dataset_id)
        wd = os.getcwd()
        os.chdir(folder_path)
        files_list = get_list_files_info("./")
        with tarfile.open(model_filename, "w:gz") as tar:
            for filename in files_list:
                tar.add(filename)

        os.chdir(wd)

        hist_filename = os.path.join(hist_folder, train_id + ".tar.gz")
        if os.path.isfile(hist_filename):
            os.remove(hist_filename)

        wd = os.getcwd()
        os.chdir(hist_folder)
        files_list = get_list_files_info("./")
        with tarfile.open(hist_filename, "w:gz") as tar:
            for filename in files_list:
                tar.add(filename)

        os.chdir(wd)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        url = f"{endpoint}/model/{dataset_id}/{train_id}"

        files = {
            'model_file': open(model_filename, 'rb'),
            'train_file': open(hist_filename, 'rb')
        }

        test_batch_filename = os.path.join(hist_folder, "test_batch-" + dataset_id + ".jpg")
        if os.path.isfile(test_batch_filename):
            files['test_batch'] = open(test_batch_filename, 'rb')

        test_augmentation_filename = os.path.join(hist_folder, "test_augmentation-" + dataset_id + ".jpg")
        if os.path.isfile(test_augmentation_filename):
            files['test_augmentation'] = open(test_augmentation_filename, 'rb')

        model_info["size"] = os.stat(model_filename).st_size
        train_info["size"] = os.stat(hist_filename).st_size
        values = {
            'model_info': json.dumps(model_info, default=str),
            'train_info': json.dumps(train_info, default=str)
        }

        response = requests.post(url, files=files, data=values, headers=msg_headers)

        if response.status_code != 201:
            raise Exception(f"Failing upload model: {response.json()}")

        os.remove(model_filename)
        os.remove(hist_filename)

        return dataset_id

    except requests.ConnectionError as error:
        log.error(f'Failing uploading model: {dataset_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing uploading model: {dataset_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing uploading model: {dataset_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_train(app_token, dataset_id, train_id, train_folder):
    try:
        log.info(f"Get train {dataset_id}-{train_id}")
        folder_path = Path(train_folder + '/' + dataset_id + '/' + train_id)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        url = f"{endpoint}/model-hist/{dataset_id}/{train_id}"
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        payload = {"download_url": True}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            log.error(f"Failing get train: {response.json()}")
            return None

        train_doc = response.json()

        dest_filename = os.path.join(str(folder_path), train_id + ".tar.gz")
        download_file(train_doc["download_url"], dest_filename)

        # expand_file
        call([
            "tar",
            "-xf", dest_filename,
            "--directory", str(folder_path)
        ])

        os.remove(dest_filename)
        return train_id

    except requests.ConnectionError as error:
        log.error(f'Failing get train_id: {train_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing get train_id: {train_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing get train_id: {train_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def insert_train_event(app_token, event):
    try:
        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        url = f"{endpoint}/train/event"

        data = {
            "event": json.dumps(event, default=str)
        }
        response = requests.post(url, data=data, headers=msg_headers)

        if response.status_code != 201:
            raise Exception(f"Failing insert event: {response.json()}")

        return True

    except requests.ConnectionError as error:
        log.error(f'Failing inserting train event. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing inserting train event. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing inserting train event - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_flow_component(app_token, flow_component_id, flow_component_folder):
    local_doc = None
    try:
        log.info(f"Get flow_component {flow_component_id}")

        folder_path = Path(flow_component_folder + '/' + flow_component_id)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        local_cache = os.path.join(flow_component_folder, flow_component_id + '.json')
        if os.path.isfile(local_cache):
            with open(local_cache) as fp:
                local_doc = json.load(fp)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        url = f"{endpoint}/flow-component/{flow_component_id}"
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        payload = {"download_url": False}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get flow-component: {response.json()}")
            return None

        flow_component_doc = response.json()
        if local_doc and flow_component_doc["version"] == local_doc["version"]:
            return local_doc

        payload = {"download_url": True}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model: {response.json()}")
            return None

        flow_component_doc = response.json()

        dest_filename = os.path.join(folder_path, flow_component_id + ".tar.gz")
        download_file(flow_component_doc["download_url"], dest_filename)

        # expand_file
        call([
            "tar",
            "-xf", dest_filename,
            "--directory", folder_path
        ])

        os.remove(dest_filename)

        if os.path.isfile(local_cache):
            os.remove(local_cache)

        with open(local_cache, 'w') as fp:
            json.dump(flow_component_doc, fp, default=str)

        return flow_component_doc

    except requests.ConnectionError as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get flow_component: {flow_component_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get flow_component: {flow_component_id}. Timeout: {error}')
        return None
    except Exception as excp:
        if local_doc:
            return local_doc

        log.error(f'Failing get flow_component: {flow_component_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_model_component(app_token, model_component_id, model_component_folder):
    local_doc = None
    try:
        log.info(f"Get model_component {model_component_id}")
        folder_path = Path(model_component_folder + '/' + model_component_id)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        local_cache = os.path.join(model_component_folder, model_component_id + '.json')
        if os.path.isfile(local_cache):
            with open(local_cache) as fp:
                local_doc = json.load(fp)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        url = f"{endpoint}/model-component/{model_component_id}"
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        payload = {"download_url": False}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model_component: {response.json()}")
            return None

        model_component_doc = response.json()
        if local_doc and model_component_doc["version"] == local_doc["version"]:
            return local_doc

        payload = {"download_url": True}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model: {response.json()}")
            return None

        model_component_doc = response.json()

        dest_filename = os.path.join(folder_path, model_component_id + ".tar.gz")
        download_file(model_component_doc["download_url"], dest_filename)

        # expand_file
        call([
            "tar",
            "-xf", dest_filename,
            "--directory", folder_path
        ])

        os.remove(dest_filename)

        if os.path.isfile(local_cache):
            os.remove(local_cache)

        with open(local_cache, 'w') as fp:
            json.dump(model_component_doc, fp, default=str)

        return model_component_id

    except requests.ConnectionError as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get model_component: {model_component_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get model_component: {model_component_id}. Timeout: {error}')
        return None
    except Exception as excp:
        if local_doc:
            return local_doc

        log.error(f'Failing get model_component: {model_component_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def clear_log(extract_path, max_files):
    files_list = [fname for fname in os.listdir(extract_path) if fname.endswith('.jpg') and not fname.endswith('_thumb.jpg')]
    if len(files_list) > max_files:
        date_list = [(filename, datetime.datetime.fromtimestamp(os.path.getmtime(os.path.join(extract_path, filename)))) for filename in files_list]
        exclude_list = sorted(date_list, key=lambda x: x[1])[:len(files_list) - max_files]
        for filename, _ in exclude_list:
            try:
                os.remove(os.path.join(extract_path, filename))
                os.remove(os.path.join(extract_path, filename[:-4] + '_data.json'))
                os.remove(os.path.join(extract_path, filename[:-4] + '_thumb.jpg'))
            except:
                pass
#----------------------------------------------------------------------------------------------------------------------------------


def generate_extract_thumbs(extract_path, thumb_size):
    """ Generate thumb image for all image files in extract folder
    """
    thumbs_list = [fname for fname in os.listdir(extract_path) if fname.endswith('_thumb.jpg')]
    for filename in os.listdir(extract_path):
        file_thumb = filename[:-4] + "_thumb.jpg"
        if filename.endswith('.jpg') and filename not in thumbs_list and file_thumb not in thumbs_list:
            img = cv2.imread(os.path.join(extract_path, filename))
            if max(img.shape) > thumb_size:
                img, _ = resize_image_scale(img, thumb_size)
            cv2.imwrite(os.path.join(extract_path, file_thumb), img)
#----------------------------------------------------------------------------------------------------------------------------------


MAX_EXTRACT_FILES = 400
THUMB_SIZE = 128
def upload_extract(app_token, dataset_id, extract_folder, max_files=MAX_EXTRACT_FILES, thumb_size=THUMB_SIZE):
    try:
        log.info(f"Upload extract {dataset_id}")
        folder_path = os.path.join(extract_folder, dataset_id)
        if not os.path.isdir(folder_path):
            raise Exception(f"Extract folder doesn't exists: {folder_path}")

        clear_log(folder_path, max_files)
        generate_extract_thumbs(folder_path, thumb_size)

        files_data = []
        file_list = [f for f in os.listdir(folder_path)]
        for filename in file_list:
            exp_id = filename[:24]
            if filename.endswith('_data.json') and (exp_id + ".jpg") in file_list and (exp_id + "_thumb.jpg") in file_list:
                try:
                    filepath = os.path.join(folder_path, filename)
                    with open(filepath, 'r') as json_file:
                        data = json.load(json_file)

                        if "date" in data:
                            data["date"] = {"$date": data["date"]}

                        if "_id" in data:
                            data["_id"] = {"$oid": data["_id"]}

                        files_data.append(data)
                except:
                    pass

        if not files_data:
            log.warning(f'Cannot upload post upload_extract: {dataset_id}. No files.')
            return dataset_id

        extract_files = {
            "files_data": files_data
        }

        with open(os.path.join(folder_path, 'extract_files.json'), 'w', newline='', encoding='utf8') as file_p:
            json.dump(extract_files, file_p, ensure_ascii=False, default=str)

        dest_filename = os.path.join(extract_folder, dataset_id + ".tar.gz")
        if os.path.isfile(dest_filename):
            os.remove(dest_filename)

        wd = os.getcwd()
        os.chdir(folder_path)
        with tarfile.open(dest_filename, "w:gz") as tar:
            for filename in os.listdir(folder_path):
                if filename.endswith('.jpg') or filename == 'extract_files.json':
                    tar.add(filename)

        os.chdir(wd)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        url = f"{endpoint}/dataset/{dataset_id}/extract"

        files = {'extract': open(dest_filename, 'rb')}
        values = {
            'dataset_id': dataset_id,
            'extract_files': extract_files
        }

        response = requests.post(url, files=files, data=values, headers=msg_headers)

        if response.status_code != 201:
            raise Exception(f"Failing upload extract files: {response.json()}")

        os.remove(dest_filename)
        return dataset_id

    except requests.ConnectionError as error:
        log.error(f'Failing post upload_extract: {dataset_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing post upload_extract: {dataset_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing post upload_extract: {dataset_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_video(app_token, video_id, video_folder):
    local_doc = None
    try:
        folder_path = Path(video_folder)
        if not folder_path.is_dir():
            folder_path.mkdir(parents=True, exist_ok=True)

        local_cache = os.path.join(video_folder, video_id + '.json')
        if os.path.isfile(local_cache):
            with open(local_cache) as fp:
                local_doc = json.load(fp)

        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        url = f"{endpoint}/video/{video_id}"
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        payload = {"download_url": False}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get video: {response.json()}")
            return None

        video_doc = response.json()
        if local_doc and video_doc["date"] == local_doc["date"]:
            return local_doc

        payload = {"download_url": True}
        response = requests.get(url, headers=msg_headers, params=payload)

        if response.status_code != 200:
            if local_doc:
                return local_doc

            log.error(f"Failing get model: {response.json()}")
            return None

        video_doc = response.json()

        dest_filename = os.path.join(video_folder, video_id + ".mp4")
        download_file(video_doc["download_url"], dest_filename)

        if os.path.isfile(local_cache):
            os.remove(local_cache)

        with open(local_cache, 'w') as fp:
            json.dump(video_doc, fp, default=str)

        return video_doc

    except requests.ConnectionError as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get video: {video_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        if local_doc:
            return local_doc

        log.error(f'Failing get video: {video_id}. Timeout: {error}')
        return None
    except Exception as excp:
        if local_doc:
            return local_doc

        log.error(f'Failing get video: {video_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def upload_video(app_token, video_id, video_file_annotation, video_file, output_file, video_folder):
    try:
        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        url = f"{endpoint}/video/{video_id}"

        video_filename = os.path.join(video_folder, video_file)
        output_filename = os.path.join(video_folder, output_file)
        files = {
            'video': open(video_filename, 'rb'),
            'output': open(output_filename, 'rb')
        }
        values = {
            "annotations": json.dumps(video_file_annotation, default=str)
        }

        response = requests.post(url, files=files, data=values, headers=msg_headers)

        if response.status_code != 201:
            raise Exception(f"Failing upload video: {response.json()}")

        return video_id

    except requests.ConnectionError as error:
        log.error(f'Failing uploading video: {video_file}-{video_id}. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing uploading video: {video_file}-{video_id}. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing uploading video: {video_file}-{video_id} - {excp}')
        return None
# ---------------------------------------------------------------------------------------------------------------------------------


def get_edge_data(app_token):
    try:
        log.info(f"Get edge_data")
        endpoint = jwt.decode(app_token, options={"verify_signature": False})['endpoint']
        msg_headers = {'Authorization' : f'Bearer {app_token}'}
        response = requests.get(f"{endpoint}", headers=msg_headers)

        if response.status_code != 200:
            log.error(f"Failing get edge_data: {response.json()}")
            return None

        return response.json()["edge_data"]

    except requests.ConnectionError as error:
        log.error(f'Failing get edge_data. Connection error: {error}')
        return None
    except requests.Timeout as error:
        log.error(f'Failing get edge_data. Timeout: {error}')
        return None
    except Exception as excp:
        log.error(f'Failing get edge_data - {excp}')
        return None
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Class to operate message queue server/client

Author: Alex Sobral de Freitas
"""

import os
import json
import datetime
import time
from typing import Dict
import pika
from bson.objectid import ObjectId

from eyeflow_sdk.log_obj import log, CONFIG
#----------------------------------------------------------------------------------------------------------------------------------

class QueueChannel(object):
    def __init__(self, queue, message_call, parent):
        self._queue = queue
        self._message_call = message_call
        self._parent = parent
        self._consumer_tag = None


    def on_channel_open(self, channel):
        log.info('Channel opened')
        self._channel = channel

        log.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

        log.info('Declaring queue %s' % self._queue)
        self._channel.queue_declare(queue=self._queue,
                                    callback=self.on_queue_declareok,
                                    durable=True)

        log.info('Adding process consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)


    def on_message(self, unused_channel, basic_deliver, properties, body):
        log.info('Received message # %s from %s: %s', basic_deliver.delivery_tag, properties.app_id, body)
        parms = json.loads(body.decode())
        self._message_call(parms)

        # Acknowledge the message delivery from RabbitMQ by sending a Basic.Ack RPC method for the delivery tag.
        self._channel.basic_ack(basic_deliver.delivery_tag)


    def on_queue_declareok(self, method_frame):
        self._consumer_tag = self._channel.basic_consume(queue=self._queue,
                                                         on_message_callback=self.on_message)


    def on_channel_closed(self, channel, reason):
        log.warning('Channel %i was closed: %s', channel, reason)
        self._parent._connection.close()


    def on_consumer_cancelled(self, method_frame):
        log.info('Consumer was cancelled remotely, shutting down: %r', method_frame)
        if self._channel:
            self._channel.close()


    def stop_consuming(self):
        if self._channel:
            log.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)


    def on_cancelok(self, unused_frame):
        log.info('RabbitMQ acknowledged the cancellation of the consumer')
        log.info('Closing the channel')
        self._channel.close()
#----------------------------------------------------------------------------------------------------------------------------------

class TopicChannel(object):
    def __init__(self, topic, route, message_call, parent):
        self._topic = topic
        self._routing_key = route
        self._queue_name = None
        self._message_call = message_call
        self._parent = parent
        self._consumer_tag = None


    def on_channel_open(self, channel):
        log.info('Channel opened')
        self._channel = channel

        log.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

        log.info('Declaring exchange %s' % self._topic)
        self._channel.exchange_declare(exchange=self._topic,
                                       callback=self.on_exchange_declareok,
                                       durable=True,
                                       exchange_type='topic')


    def on_exchange_declareok(self, unused_frame):
        log.info('Exchange declared')
        log.info('Declaring queue')
        self._channel.queue_declare(queue='',
                                    callback=self.on_queue_declareok,
                                    exclusive=True,
                                    auto_delete=True)


    def on_queue_declareok(self, method_frame):
        log.info('Queue declare ok: consumer_count={}, message_count={}, queue={}'.format(method_frame.method.consumer_count, method_frame.method.message_count, method_frame.method.queue))
        self._queue_name = method_frame.method.queue

        log.info('Binding %s to %s with %s' % (self._topic, self._queue_name, self._routing_key))
        self._channel.queue_bind(exchange=self._topic,
                                 routing_key=self._routing_key,
                                 queue=self._queue_name,
                                 callback=self.on_bindok)


    def on_bindok(self, unused_frame):
        log.info('Queue bound')
        log.info('Adding process consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)
        self._consumer_tag = self._channel.basic_consume(queue=self._queue_name,
                                                         on_message_callback=self.on_message)


    def on_message(self, unused_channel, basic_deliver, properties, body):
        event_parms = json.loads(body.decode())
        queue_parms = {"timestamp": datetime.datetime.now().isoformat(),
                       "routing_key": basic_deliver.routing_key,
                       "app_id": properties.app_id
        }
        self._message_call(queue_parms, event_parms)

        # Acknowledge the message delivery from RabbitMQ by sending a Basic.Ack RPC method for the delivery tag.
        self._channel.basic_ack(basic_deliver.delivery_tag)


    def on_channel_closed(self, channel, reason):
        log.warning('Channel %i was closed: %s', channel, reason)
        self._parent._connection.close()


    def on_consumer_cancelled(self, method_frame):
        log.info('Consumer was cancelled remotely, shutting down: %r', method_frame)
        if self._channel:
            self._channel.close()


    def stop_consuming(self):
        if self._channel:
            log.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            self._channel.basic_cancel(self.on_cancelok, self._consumer_tag)


    def on_cancelok(self, unused_frame):
        log.info('RabbitMQ acknowledged the cancellation of the consumer')
        log.info('Closing the channel')
        self._channel.close()
#----------------------------------------------------------------------------------------------------------------------------------

class QueueConsumer(object):
    def __init__(self, amqp_url, queues=None, topic=None):
        self._connection = None
        self._channels = []
        self._queues = queues
        self._topic = topic
        self._connection_tries = 0
        self._closing = False
        self._url = amqp_url


    def connect(self):
        log.info('Connecting to %s', self._url)
        self._connection = pika.SelectConnection(pika.URLParameters(self._url),
                                                 on_open_callback=self.on_connection_open,
                                                 on_open_error_callback=self.on_connection_error)


    def on_connection_error(self, unused_connection, excpt):
        if self._connection_tries < 5:
            log.warning('Fail to connect - {}. Wait 5 seconds'.format(repr(excpt)))
            time.sleep(5)
            self._connection_tries += 1
            self.connect()
            self._connection.ioloop.start()
        else:
            log.error('Fail to connect to broker')
            exit(1)


    def on_connection_open(self, unused_connection):
        log.info('Connection opened')
        log.info('Adding connection close callback')
        self._connection.add_on_close_callback(self.on_connection_closed)

        if self._queues is not None:
            for queue in self._queues:
                log.info('Creating queue channel: ' + queue[0])
                new_channel = QueueChannel(queue[0], queue[1], self)
                self._channels.append(new_channel)
                self._connection.channel(on_open_callback=new_channel.on_channel_open)
        elif self._topic is not None:
            log.info('Creating topic channel: ' + self._topic[0])
            new_channel = TopicChannel(self._topic[0], self._topic[1], self._topic[2], self)
            self._channels.append(new_channel)
            self._connection.channel(on_open_callback=new_channel.on_channel_open)


    def on_connection_closed(self, connection, reason):
        self._channels = []
        if self._closing:
            self._connection.ioloop.stop()
        else:
            log.warning('Connection closed, reopening in 5 seconds: %s', reason)
            self._connection.ioloop.call_later(5, self.reconnect)


    def reconnect(self):
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self._closing:
            # Create a new connection
            self.connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()


    def run(self):
        self._connection_tries = 0
        self.connect()
        self._connection.ioloop.start()


    def stop(self):
        log.info('Stopping')
        self._closing = True
        for channel in self._channels:
            channel.stop_consuming()

        self._connection.ioloop.stop()
        log.info('Stopped')


    def close_connection(self):
        log.info('Closing connection')
        self._connection.close()
#----------------------------------------------------------------------------------------------------------------------------------

def publish_message(message, queue, topic='') -> None:
    """
    Publish a message to a queue or a exchange
    message: dict - Dict with field-value of the message
    queue: str - Name of the queue to publish the message
    topic: str - Topic to publish, leave blank to publish to a queue
    """
    if "MQ_URL" in os.environ:
        mq_url = os.environ["MQ_URL"]
    elif "mq-service" in CONFIG:
        mq_parms = CONFIG["mq-service"]
        mq_url = f"amqp://{mq_parms['user']}:{mq_parms['pass']}@{mq_parms['host']}:{mq_parms['port']}"
    else:
        raise Exception('Need parms of connection os.environ["MQ_URL"] or CONFIG["mq-service"]')

    mq_connection = pika.BlockingConnection(pika.URLParameters(mq_url))
    mq_channel = mq_connection.channel()

    mq_channel.basic_publish(
        exchange=topic,
        routing_key=queue,
        properties=pika.BasicProperties(content_type="application/json"),
        body=json.dumps(message, default=str)
    )

    mq_channel.close()
    mq_connection.close()
#----------------------------------------------------------------------------------------------------------------------------------


def send_message(message, queue, topic='', timeout=5) -> Dict:
    """
    Send a message to a queue or a exchange and return response to caller
    message: dict - Dict with field-value of the message
    queue: str - Name of the queue to publish the message
    topic: str - Topic to publish, leave blank to publish to a queue
    timeout: int - Timeout in seconds
    """

    if "MQ_URL" in os.environ:
        mq_url = os.environ["MQ_URL"]
    elif "mq-service" in CONFIG:
        mq_parms = CONFIG["mq-service"]
        mq_url = f"amqp://{mq_parms['user']}:{mq_parms['pass']}@{mq_parms['host']}:{mq_parms['port']}"
    else:
        raise Exception('Need parms of connection os.environ["MQ_URL"] or CONFIG["mq-service"]')

    mq_connection = pika.BlockingConnection(pika.URLParameters(mq_url))
    mq_channel = mq_connection.channel()

    corr_id = None
    response = None

    def on_response(channel, basic_deliver, properties, body):
        nonlocal response
        if corr_id == properties.correlation_id:
            response = json.loads(body.decode())

    result = mq_channel.queue_declare(queue='', exclusive=True)
    response_queue_name = result.method.queue

    mq_channel.basic_consume(
        on_message_callback=on_response,
        queue=response_queue_name,
        auto_ack=True
    )

    corr_id = str(ObjectId())

    mq_channel.basic_publish(
        exchange=topic,
        routing_key=queue,
        properties=pika.BasicProperties(
            content_type="application/json",
            reply_to=response_queue_name,
            correlation_id=corr_id,
        ),
        body=json.dumps(message, default=str)
    )

    start_timeout = datetime.datetime.now()
    while not response:
        mq_connection.process_data_events()
        time.sleep(0.01)
        if (datetime.datetime.now() - start_timeout).total_seconds() > timeout:
            raise Exception("Timeout waiting response")

    return response
#----------------------------------------------------------------------------------------------------------------------------------



"""
    Example code for resetting the USB port that a Teensy microcontroller is
    attached to. There are a lot of situations where a Teensy or Arduino can
    end up in a bad state and need resetting, this code is useful for
"""

import os
import fcntl
import subprocess
#----------------------------------------------------------------------------------------------------------------------------------

# Equivalent of the _IO('U', 20) constant in the linux kernel.
USBDEVFS_RESET = ord('U') << (4*2) | 20

def get_realsense():
    """
        Gets the devfs path to a Realsense Camera by scraping the output
        of the lsusb command

        The lsusb command outputs a list of USB devices attached to a computer
        in the format:
            Bus 002 Device 009: ID 16c0:0483 Intel Corp.
        The devfs path to these devices is:
            /dev/bus/usb/<busnum>/<devnum>
        So for the above device, it would be:
            /dev/bus/usb/002/009
        This function generates that path.
    """
    proc = subprocess.Popen(['lsusb'], stdout=subprocess.PIPE)
    out = proc.communicate()[0].decode()
    lines = out.split('\n')
    for line in lines:
        if 'Intel Corp' in line:
            parts = line.split()
            bus = parts[1]
            dev = parts[3][:3]
            return '/dev/bus/usb/%s/%s' % (bus, dev)
#----------------------------------------------------------------------------------------------------------------------------------

def send_reset(dev_path):
    """
        Sends the USBDEVFS_RESET IOCTL to a USB device.

        dev_path - The devfs path to the USB device (under /dev/bus/usb/)
                   See get_teensy for example of how to obtain this.
    """
    fd = os.open(dev_path, os.O_WRONLY)
    try:
        fcntl.ioctl(fd, USBDEVFS_RESET, 0)
    finally:
        os.close(fd)
#----------------------------------------------------------------------------------------------------------------------------------

def reset_realsense():
    """
        Finds a teensy and reset it.
    """
    send_reset(get_realsense())
#----------------------------------------------------------------------------------------------------------------------------------



"""
SiliconLife Eyeflow
Dataset manipulation functions

Author: Alex Sobral de Freitas
"""

import os
import json
import random
import copy
import datetime
import pytz
import dateutil.parser
import requests

import cv2
import numpy as np
import h5py

from bson.objectid import ObjectId

from eyeflow_sdk.log_obj import log, CONFIG
import eyeflow_sdk.edge_client as edge_client
#----------------------------------------------------------------------------------------------------------------------------------

def default_json_converter(obj):
    if isinstance(obj, datetime.datetime):
        return obj.replace(tzinfo=datetime.timezone.utc).isoformat()
    else:
        return str(obj)
#----------------------------------------------------------------------------------------------------------------------------------


class Dataset():
    """
    Class to serialize and deserialize datasets to/from a file
    """
    def __init__(self, dataset_id, app_token=None):
        self.id = str(ObjectId(dataset_id))
        self.dataset_name = None
        self._app_token = app_token

        self.parms = {}
        self.examples = []
        self.images = {}


    @staticmethod
    def get_dataset_default_parms():
        """
        Returns default dataset parms
        """
        filename = os.path.join(os.path.dirname(__file__), 'dataset_default_parms.json')
        with open(filename, 'r', newline='', encoding='utf8') as fp:
            default_parms = json.load(fp)

        return default_parms


    @staticmethod
    def get_data_augmentation_default_parms():
        """
        Returns default data augmentation parms
        """
        filename = os.path.join(os.path.dirname(__file__), 'data_augmentation_default_parms.json')
        if not os.path.isfile(filename):
            return None

        with open(filename, 'r', newline='', encoding='utf8') as fp:
            default_parms = json.load(fp)

        return default_parms


    @staticmethod
    def get_dataset_types():
        """
        Returns dataset types
        """
        filename = os.path.join(os.path.dirname(__file__), 'dataset_types.json')
        if not os.path.isfile(filename):
            return None

        with open(filename, 'r', newline='', encoding='utf8') as fp:
            default_parms = json.load(fp)

        return default_parms


    def update_default_parms(self):
        """
        Performs an update in parms with all default parms read from database
        """
        train_parms = {}
        dnn_parms = {}
        dataset_default_parms = Dataset.get_dataset_default_parms()
        if dataset_default_parms:
            train_parms = dataset_default_parms["default_parms"][self.parms["info"]["type"]]["train_parms"]
            dnn_parms = dataset_default_parms["default_parms"][self.parms["info"]["type"]]["dnn_parms"]

        input_shape = None
        if "dnn_parms" in self.parms:
            dnn_parms.update(self.parms["dnn_parms"])
            if "input_shape" in self.parms["dnn_parms"]:
                input_shape = copy.deepcopy(self.parms["dnn_parms"]["input_shape"])

            if "component" in self.parms["dnn_parms"] and "component_id" not in self.parms["dnn_parms"]:
                if self.parms["dnn_parms"]["component"] == "objdet_af":
                    dnn_parms["component_id"] = "6143a1faef5cc63fd4c177b1"
                elif self.parms["dnn_parms"]["component"] == "objdet":
                    dnn_parms["component_id"] = "6143a1edef5cc63fd4c177b0"
                elif self.parms["dnn_parms"]["component"] == "class_cnn":
                    dnn_parms["component_id"] = "614388073a692cccdab0e69b"
                elif self.parms["dnn_parms"]["component"] == "obj_location":
                    dnn_parms["component_id"] = "6178516681cbe716153175b0"

        elif "network_parms" in self.parms and "dnn_parms" in self.parms["network_parms"]:
            dnn_parms.update(self.parms["network_parms"]["dnn_parms"])

        if "train_parms" in self.parms:
            train_parms.update(self.parms["train_parms"])
            if not input_shape:
                input_shape = copy.deepcopy(self.parms["train_parms"]["input_shape"])
            if "input_shape" in train_parms:
                del train_parms["input_shape"]
        elif "network_parms" in self.parms:
            train_parms.update(self.parms["network_parms"])
            if "dnn_parms" in train_parms:
                del train_parms["dnn_parms"]

            if "input_shape" in train_parms:
                del train_parms["input_shape"]

        self.parms["train_parms"] = train_parms
        self.parms["dnn_parms"] = dnn_parms
        self.parms["dnn_parms"]["input_shape"] = input_shape

        if "network_parms" in self.parms:
            del self.parms["network_parms"]

        data_augmentation_parms = {}
        data_augmentation_default_parms = Dataset.get_data_augmentation_default_parms()
        if data_augmentation_default_parms:
            data_augmentation_parms = data_augmentation_default_parms["default_parms"]
            data_augmentation_parms.update(self.parms.get("data_augmentation_parms", {}))

        self.parms["data_augmentation_parms"] = data_augmentation_parms


    def load_data(self):
        """
        Load dataset data from ws
        """
        if not self._app_token:
            raise Exception('AppToken not set')

        dataset = edge_client.get_dataset(self._app_token, self.id)
        if not dataset:
            raise Exception(f'Fail loading dataset_id {self.id}')

        self.dataset_name = dataset["dataset_parms"].get("name")
        self.parms = dataset["dataset_parms"]
        self.examples = dataset["annotations"]
        self.update_default_parms()
        log.info(f"Load dataset from database {len(self.examples)} examples")


    @staticmethod
    def load_file_from_cloud(url):
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            img = b""
            for chunk in r.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment 'if' and set chunk_size parameter to None.
                #if chunk:
                img += chunk

            return img


    def load_image(self, example_img):
        """
        Load image to memory from disk or cloud
        """
        for exp in self.examples:
            if exp["example"] == example_img:
                break
        else:
            raise Exception(f"Example not found {example_img}")

        exp_folder = os.path.join(CONFIG["file-service"]["dataset"]["local_folder"], self.id)
        if os.path.isfile(os.path.join(exp_folder, exp["example"])):
            with open(os.path.join(exp_folder, exp["example"]), 'rb') as fp:
                self.images[exp["example"]] = fp.read()
        else:
            self.images[exp["example"]] = self.load_file_from_cloud(exp["download_url"])


    def load_all_images(self):
        """
        Load images
        """
        for exp in self.examples:
            if exp["example"] not in self.images:
                exp_folder = os.path.join(CONFIG["file-service"]["dataset"]["local_folder"], self.id)
                if os.path.isfile(os.path.join(exp_folder, exp["example"])):
                    with open(os.path.join(exp_folder, exp["example"]), 'rb') as fp:
                        self.images[exp["example"]] = fp.read()
                else:
                    self.images[exp["example"]] = self.load_file_from_cloud(exp["download_url"])


    def save_images_to_disk(self):
        """
        Save images to disk
        """
        self.load_all_images()
        exp_folder = os.path.join(CONFIG["file-service"]["dataset"]["local_folder"], self.id)

        for exp in self.examples:
            if os.path.isfile(os.path.join(exp_folder, exp["example"])):
                os.remove(os.path.join(exp_folder, exp["example"]))

            with open(os.path.join(exp_folder, exp["example"]), 'wb') as fp:
                fp.write(self.images[exp["example"]])


    def get_example_img(self, example_img):
        """
        Returns an image in opencv format
        """
        if example_img not in self.images:
            self.load_image(example_img)

        return cv2.imdecode(np.frombuffer(self.images[example_img], dtype=np.uint8), cv2.IMREAD_COLOR)


    def get_train_subsets(self, shuffle=True):
        """
        Partitioning of dataset in 3 groups: train, validation and test
        """
        if not self.parms:
            raise Exception('Must load or import dataset first')

        train_dataset = Dataset(self.id, app_token=self._app_token)
        train_dataset.parms = copy.deepcopy(self.parms)

        val_dataset = Dataset(self.id, app_token=self._app_token)
        val_dataset.parms = copy.deepcopy(self.parms)

        test_dataset = Dataset(self.id, app_token=self._app_token)
        test_dataset.parms = copy.deepcopy(self.parms)

        if shuffle:
            random.shuffle(self.examples)

        val_size = int(len(self.examples) * self.parms["train_parms"]["val_size"])
        if val_size < 1:
            val_size = 1

        test_size = int(len(self.examples) * self.parms["train_parms"]["test_size"])
        if test_size < 1:
            test_size = 1

        if (val_size + test_size) > len(self.examples):
            raise Exception("Insufficient examples to train: %d" % len(self.examples))

        train_dataset.examples = copy.deepcopy(self.examples[test_size + val_size:])
        val_dataset.examples = copy.deepcopy(self.examples[test_size:test_size + val_size])
        test_dataset.examples = copy.deepcopy(self.examples[:test_size])

        return train_dataset, val_dataset, test_dataset


    @staticmethod
    def export_to_hdf5(export_data, parms, examples, images, filename, dataset_id):
        """
        Export dataset to file in hdf5 format
        """
        example_list = []
        image_list = []
        for exp in examples:
            example_list.append(json.dumps(exp, ensure_ascii=False, default=default_json_converter))
            image_list.append(images[exp["example"]])

        dt = h5py.special_dtype(vlen=str)
        example_list = np.array(example_list, dtype=dt)
        image_list = np.array(image_list)

        export_path = os.path.join(CONFIG["file-service"]["export"]["local_folder"], dataset_id)

        with h5py.File(os.path.join(export_path, filename), 'w') as dsetfile:
            dsetfile.create_dataset('export_data', data=json.dumps(export_data, ensure_ascii=False, default=default_json_converter), dtype=dt)
            dsetfile.create_dataset('examples', data=example_list, compression="gzip", compression_opts=9)
            dsetfile.create_dataset('images', data=image_list)
            dsetfile.create_dataset('parms', data=json.dumps(parms, ensure_ascii=False, default=default_json_converter), dtype=dt)


    def export_dataset(self, filename=None):
        """
        Export dataset to folder adding only differences if a base export exists
        """

        if not filename:
            filename = self.id + ".dset"

        self.load_all_images()

        export_data = {
            "export_version": "2",
            "export_date": pytz.utc.localize(datetime.datetime.now()),
            "dataset": self.dataset_name,
            "dataset_id": self.id,
            "num_examples": len(self.examples),
            "images_list": list(self.images.keys())
        }

        # Dataset.export_to_file(filename)
        Dataset.export_to_hdf5(export_data, self.parms, self.examples, self.images, filename, self.id)
        # if self._app_token is not None:

        log.info(f"Dataset exported to file {len(self.examples)} examples")


    @staticmethod
    def import_from_hdf5(filename, dataset_id, retrieve_images=True):
        """
        Import dataset from file in protobuf format
        """
        def convert_date(str_date):
            return dateutil.parser.isoparse(str_date).replace(tzinfo=None)

        export_path = os.path.join(CONFIG["file-service"]["export"]["local_folder"], dataset_id)
        with h5py.File(os.path.join(export_path, filename), 'r') as dsetfile:
            export_data = json.loads(dsetfile['export_data'][()])

            example_list = dsetfile['examples'][()].tolist()
            image_list = dsetfile['images'][()].tolist()
            parms = dsetfile['parms'][()]


        parms = json.loads(parms)
        parms["_id"] = ObjectId(dataset_id)
        parms["info"]["creation_date"] = convert_date(parms["info"].get("creation_date"))
        parms["info"]["modified_date"] = convert_date(parms["info"].get("modified_date"))

        examples = []
        for exp in example_list:
            exp = json.loads(exp)
            exp["_id"] = ObjectId()
            exp["date"] = convert_date(exp.get("date"))
            exp["dataset_id"] = ObjectId(dataset_id)
            if "modified_date" in exp:
                exp["modified_date"] = convert_date(exp.get("modified_date"))

            examples.append(exp)

        if not retrieve_images:
            return parms, examples

        images = {}
        for exp, img in zip(examples, image_list):
            images[exp["example"]] = img

        return export_data, parms, examples, images


    def import_dataset(self, filename=None):
        """
        Import dataset from folder with base and diffs
        """

        if not filename:
            if self.id:
                filename = self.id + ".dset"
            elif self.dataset_name:
                filename = self.dataset_name + ".dset"
            else:
                raise Exception(f'Must define import filename')

        export_data, self.parms, self.examples, self.images = Dataset.import_from_hdf5(filename, self.id)

        self.dataset_name = str(self.parms["name"])
        self.id = str(self.parms["_id"])

        if len(self.examples) != export_data["num_examples"]:
            raise Exception(f"Invalid number of examples. Imported: {len(self.examples)} - Exported: {export_data['num_examples']}")

        if len(self.images) != len(export_data["images_list"]):
            raise Exception(f"Invalid number of images. Imported: {len(self.images)} - Exported: {len(export_data['images_list'])}")

        log.info("Dataset imported from file %d examples" % len(self.examples))
# ---------------------------------------------------------------------------------------------------------------------------------
